use std::io::{self, ErrorKind, Read, Write};

use thiserror::Error;

#[derive(Clone, Copy)]
pub struct Header {
    pub bytes: [u8; 3],
    pub version: u8
}


#[derive(Error, Debug)]
pub enum ProtocolWriterError {
    #[error("error using underlying writer")]
    Io(#[from] io::Error)
}

pub type ProtocolWriterResult<T> = Result<T, ProtocolWriterError>;

pub struct ProtocolWriter<W: Write> {
    writer: W
}

#[derive(Error, Debug)]
pub enum ProtocolReaderError {
    // this is explicitly supported: rxi images are progressive
    #[error("EOF before end of image")]  
    EarlyEof,

    #[error("wrong header")]
    WrongHeader,

    #[error("header has unsupported version")]
    WrongVersion,

    #[error("error using underlying reader")]
    Io(#[from] io::Error),
}

pub type ProtocolReaderResult<T> = Result<T, ProtocolReaderError>;

pub struct ProtocolReader<R: Read> {
    reader: R
}

impl<W: Write> ProtocolWriter<W> {
    pub fn new(writer: W) -> Self {
        Self { writer }
    }

    pub fn destroy(self) -> W {
        self.writer
    }

    pub fn write_header(&mut self, value: Header) -> ProtocolWriterResult<()> {
        self.writer.write_all(&value.bytes)?;
        self.writer.write_all(&[value.version])?;
        Ok(())
    }
    pub fn write_u32_wide(&mut self, value: u32) -> ProtocolWriterResult<()> {
        self.writer.write_all(&value.to_le_bytes())?;
        Ok(())
    }
    pub fn write_i32_packed(&mut self, value: i32) -> ProtocolWriterResult<()> {
        // We reserve i8::MAX and i16::MAX as signal values that `value` is too big
        // for each respective type
        if (i8::MIN as i32..i8::MAX as i32).contains(&value) {
            self.writer.write_all(&(value as i8).to_le_bytes())?;
            return Ok(())
        } 
        self.writer.write_all(&(i8::MAX).to_le_bytes())?;

        if (i16::MIN as i32..i16::MAX as i32).contains(&value) {
            self.writer.write_all(&(value as i16).to_le_bytes())?;
            return Ok(())
        } 
        self.writer.write_all(&(i16::MAX).to_le_bytes())?;
        self.writer.write_all(&value.to_le_bytes())?;
        Ok(())
    }
}

impl<R: Read> ProtocolReader<R> {
    pub fn new(reader: R) -> Self {
        Self { reader }
    }

    pub fn read_header(&mut self) -> ProtocolReaderResult<Header> {
        let mut bytes_buf = [0; 3];
        self.read_exact(&mut bytes_buf)?;

        let mut version_buf = [0; 1];
        self.read_exact(&mut version_buf)?;

        Ok(Header {
            bytes: bytes_buf,
            version: version_buf[0],
        })
    }

    pub fn read_u32_wide(&mut self) -> ProtocolReaderResult<u32> {
        let mut u32_buf = [0; 4];
        self.read_exact(&mut u32_buf)?;

        Ok(u32::from_le_bytes(u32_buf))
    }

    pub fn read_i32_packed(&mut self) -> ProtocolReaderResult<i32> {
        let mut i8_buf = [0; 1];
        self.read_exact(&mut i8_buf)?;
        let i8_value = i8::from_le_bytes(i8_buf);
        if i8_value != i8::MAX { return Ok(i8_value as i32) }

        let mut i16_buf = [0; 2];
        self.read_exact(&mut i16_buf)?;
        let i16_value = i16::from_le_bytes(i16_buf);
        if i16_value != i16::MAX { return Ok(i16_value as i32) }

        let mut i32_buf = [0; 4];
        self.read_exact(&mut i32_buf)?;
        let i32_value = i32::from_le_bytes(i32_buf);
        Ok(i32_value)
    }

    // wrap UnexpectedEof, since that's an error for other Io clients,
    // but it's specifically not an error for us
    fn read_exact(&mut self, buf: &mut [u8]) -> ProtocolReaderResult<()> {
        match self.reader.read_exact(buf) {
            Ok(_) => Ok(()),
            Err(e) => {
                if e.kind() == ErrorKind::UnexpectedEof {
                    return Err(ProtocolReaderError::EarlyEof)
                }
                return Err(ProtocolReaderError::Io(e));
            }
        }
    }
}